########################################################################
#
# File Name:            WebDAVHandler.py
#
# Documentation:
#
"""
Basic read-only WebDAV server which uses pyDAV to serve XmlServer content to WebDAV clients
WWW: http://4suite.org         e-mail: support@4suite.org

Copyright (c) 2000 Fourthought, Inc. USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""
import re, urlparse, time, sys, urllib, os, string

from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Common import ResourceTypes

import CommonHandler


from DAV.propfind import PROPFIND
from DAV.delete import DELETE
from DAV.davcopy import COPY
from DAV.davmove import MOVE
from DAV.errors import *
from DAV.ResourceTypes import *
from DAV.status import STATUS_CODES


class FourSSXmlServerDAVProxy:

    """
    Implements pyDAV's dav_interface for a running XmlServer instance
    """

    PROPS={"DAV:" : ('creationdate',
                     'displayname',
                     'getcontentlanguage',
                     'getcontentlength',
                     'getcontenttype',
                     'getetag',
                     'getlastmodified',
                     'lockdiscovery',
                     'resourcetype',
                     'source',
                     'supportedlock'),
           "NS2" : ("p1","p2")
           }

    M_NS={"DAV:" : "_get_dav",
          "NS2"  : "ns2" }


    def __init__(self):
        self._uris = {}


    def get_propnames(self,uri,*args):
        return self.PROPS


    def get_prop(self,uri,ns,propname,*args):
        if self.M_NS.has_key(ns):
            prefix=self.M_NS[ns]
        else:
            raise DAV_NotFound
        mname=prefix+"_"+propname
        if not hasattr(self,mname):
            #print "Missing Function %s" % mname
            raise DAV_NotFound
        try:
                m=getattr(self,mname)
                r=apply(m,(uri,) + args)
                return r
        except FtServerServerException, e:
            if e.code == FtServerIf.PERMISSION_DENIED:
                raise DAV_Secret
            raise


    def is_collection(self,uri,host):
        """ return type of object """
        try:
            rt = self._getResourceInformation(host,uri)
        except:
            return 0
        return rt.isResourceType(ResourceTypes.ResourceType.CONTAINER)


    def exists(self,uri,*args):
        try:
            rt = self._getResourceInformation(args[0],uri)
            return 1
        except:
            return 0



    def get_childs(self,uri,*args):
        """ return the child objects as URIs for the given URI """
        return self._do_safe(self._get_childs,(uri,)+args)

    def _get_childs(self,uri,*args):
        ri = self._getResourceInformation(args[0],uri)
        if not ri.isResourceType(ResourceTypes.ResourceType.CONTAINER):
            return []
        cont = self._fetchContainer(args[0],uri)
        return map(lambda x:x.uri,cont)


    def get_data(self,uri,*args):
        """ return the content of an object """
        return self._do_safe(self._get_data,(uri,)+args)

    def _get_data(self,uri,*args):
        rt = self._getResourceInformation(args[0],uri)
        if rt.isResourceType(ResourceTypes.ResourceType.CONTAINER):
            raise DAV_NotFound
        else:
            body,imt = self._fetchDocument(args[0],uri)
            return body

    def mkcol(self,uri,*args):
        """ create a new collection """
        return self._do_safe(self._mkcol,(uri,)+args)

    def _mkcol(self,uri,*args):
        args[0].createContainer(uri,0)
        return 201


    def rmcol(self,uri,*args):
        """ delete a collection """
        return self._do_safe(self._rmcol,(uri,)+args)

    def _rmcol(self,uri,*args):
        try:
            args[0].deleteResource(uri)
        except DAV_Error, e:
            return {uri:e.args[0]}

        self._uris[uri] = [None,None]
        return {}

    def rm(self,uri,*args):
        """ delete a normal resource, raise appropriate errors """
        return self._do_safe(self._rm,(uri,)+args)

    def _rm(self,uri,*args):
        try:
            args[0].deleteResource(uri)
        except DAV_Error, e:
            return {uri:e.args[0]}

        self._uris[uri] = [None,None]
        return None

    def put(self,uri,data,content_type,*args):
        return self._do_safe(self._put,(uri,data,content_type)+args)

    def _put(self,uri,data,content_type,*args):
        """ put the object into the filesystem """

        if not args[0].hasResource(uri):
            #FIXME add some sort of mapping from extension/mime/type to DOCDEF name
            if content_type in ['text/xml' or 'application/xml']:
                args[0].createDocument('BASE_XML',uri,data)
            else:
                args[0].createRawFile(uri,content_type,data)

        else:
            doc = args[0].fetchFile(uri)
            doc.setContent(data)
            self._uris[uri] = [None,None]
        return None

    def copyone(self,src,dest,overwrite,*args):
        return self._do_safe(self._copyone,(src,dest,overwrite)+args)

    def _copyone(self,src,dest,overwrite,*args):


        if args[0].hasResource(dest):
            if overwrite:
                apply(self._rm,(dest,)+args)
            else:
                raise DAV_Error, 412
        if not args[0].hasResource(src):
            raise DAV_NotFound
        body,imt = self._fetchDocument(args[0],src)
        apply(self._put,(dest,body,imt) + args)
        return None

    def copytree(self,src,dest,overwrite,*args):
        return self._do_safe(self._copytree,(src,dest,overwrite)+args)

    def _copytree(self,src,dest,overwrite,*args):
        raise DAV_Forbidden



    def moveone(self,src,dest,overwrite,*args):
        return self._do_safe(self._moveone,(src,dest,overwrite)+args)

    def _moveone(self,src,dest,overwrite,*args):

        if args[0].hasResource(dest):
            if overwrite:
                apply(self._rm,(dest,)+args)
            else:
                raise DAV_Error, 412
        if not args[0].hasResource(src):
            raise DAV_NotFound
        body,imt = self._fetchDocument(args[0],src)
        apply(self._put,(dest,body,imt) + args)
        apply(self._rm,(src,)+args)
        return None


    def movetree(self,src,dest,overwrite,*args):
        return self._do_safe(self._movetree,(src,dest,overwrite)+args)

    def _movetree(self,src,dest,overwrite,*args):
        raise DAV_Forbidden


    #Property callbacks
    def _get_dav_creationdate(self,uri,*args):
        """ return the last modified date of the object """
        return self._do_safe(self.__get_dav_creationdate,(uri,)+args)

    def __get_dav_creationdate(self,uri,*args):
        ri = self._getResourceInformation(args[0],uri)
        date=ri.creationDate
        return time.strftime("%Y-%m-%dT%H-%M-%SZ",time.localtime(date))

    def _get_dav_displayname(self,uri,*args):
        return self._do_safe(self.__get_dav_displayname,(uri,)+args)

    def __get_dav_displayname(self,uri,*args):
        rt = self._getResourceInformation(args[0],uri)
        uparts=urlparse.urlparse(uri)
    fileloc=uparts[2]
        if fileloc == '/':
            return '/'
        return os.path.basename(fileloc)


    def _get_dav_getcontentlength(self,uri,*args):
        return self._do_safe(self.__get_dav_getcontentlength,(uri,)+args)

    def __get_dav_getcontentlength(self,uri,*args):
        """ return the content length of an object """
        rt = self._getResourceInformation(args[0],uri)
        if rt.isResourceType(ResourceTypes.ResourceType.CONTAINER):
            return 0
        else:
            return rt.size

    def _get_dav_getcontenttype(self,uri,*args):
        return self._do_safe(self.__get_dav_getcontenttype,(uri,)+args)

    def __get_dav_getcontenttype(self,uri,*args):
        """ returns the content type of an object """
        rt = self._getResourceInformation(args[0],uri)
        if rt.isResourceType(ResourceTypes.ResourceType.CONTAINER):
            return "text/html"
        else:
            print uri, rt.imt
            return rt.imt


    def _get_dav_resourcetype(self,uri,*args):
        return self._do_safe(self.__get_dav_getresourcetype,(uri,)+args)

    def __get_dav_getresourcetype(self,uri,*args):
        """ return type of object """
        rt = self._getResourceInformation(args[0],uri)
        if rt.isResourceType(ResourceTypes.ResourceType.CONTAINER):
            return COLLECTION
        else:
            return OBJECT

    def _get_dav_getlastmodified(self,uri,*args):
        return self._do_safe(self.__get_dav_getlastmodified,(uri,)+args)

    def __get_dav_getlastmodified(self,uri,*args):
        """ return the last modified date of the object """

        ri = self._getResourceInformation(args[0],uri)
        date=ri.lastModifiedDate
        return time.strftime("%a, %d %b %Y %H:%M:%S %Z",time.localtime(date))

    def _get_dav_getetag(self,uri,*args):
        return uri #???

    def _get_dav_getcontentlanguage(self,uri,*args):
        return 'en'


    #Front Page callbacks
    def _get_dav_isreadonly(self,uri,*args):
        return self._do_safe(self.__get_dav_isreadonly,(uri,)+args)

    def __get_dav_isreadonly(self,uri,*args):
        """ return the last modified date of the object """
        return '0'


    _get_dav_lastaccessed =  _get_dav_getlastmodified
    _get_dav_name = _get_dav_displayname


    def _get_dav_parentname(self,uri,*args):
        return self._do_safe(self.__get_dav_parentname,(uri,)+args)

    def __get_dav_parentname(self,uri,*args):
        rt = self._getResourceInformation(args[0],uri)
        uparts=urlparse.urlparse(uri)
    fileloc=uparts[2]
        if fileloc == '/':
            return ''
        return os.path.dirname(fileloc)

    def _get_dav_href(self,uri,*args):
        return uri

    def _get_dav_ishidden(self,uri,*args):
        return '0'

    def _get_dav_iscollection(self,uri,*args):
        return str(apply(self._get_dav_resourcetype,(uri,)+args) == COLLECTION)

    def _get_dav_isroot(self,uri,*args):
        return str(apply(self._get_dav_displayname,(uri,)+args) == '/')



    #Some helpers:
    def _getResourceInformation(self,host,uri):
        ri,data = self._uris.get(uri,(None,None))
        if ri:
            return ri
        ri = host.fetchResourceInformation(uri)
        while ri.resourceType == ResourceTypes.ResourceType.ALIAS:
            ri = host.fetchResourceInformation(ri.referenceUri)
        self._uris[uri] = [ri,data]
        return ri

    def _fetchDocument(self,host,uri):
        self._getResourceInformation(host,uri)
        ri,data = self._uris.get(uri,(None,None))
        if data:
            return data
        doc = host.fetchResource(uri)
        if doc.isResourceType(ResourceTypes.ResourceType.FILE):
            data = (doc.getContent(),doc.getImt())
        self._uris[uri] = [ri,data]
        return data

    def _fetchContainer(self,host,uri):
        ri,data = self._uris.get(uri,(None,None))
        if data:
            return data
        data = host.fetchContainer(uri)
        self._uris[uri] = [ri,data]
        return data

    def _do_safe(self,func,args):
        try:
            return apply(func,args)
        except FtServerServerException, e:
            if e.code == Error.UNKNOWN_URI:
                raise DAV_NotFound
            if e.code == Error.INVALID_URI:
                raise DAV_NotFound
            if e.code == Error.PERMISSION_DENIED:
                raise DAV_Forbidden
            raise







class FssDavHandler(CommonHandler.CommonHandler):

    def __init__(self, interfaceClass, documentRoot, authRealm):
        self.interfaceClass = interfaceClass
        self.documentRoot = documentRoot
        self.authenticationRealm = authRealm

    def do_PROPFIND(self, request):

        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error
        try:

            host = self._getHost(repo, request)

            # which Depth?
            if request.headers.has_key('Depth'):
                d=request.headers['Depth']
            else:
                d="infinity"

            ic = self.interfaceClass()
            uri=urllib.basejoin(host.getUri(),request.uri)
            uri=urllib.unquote(uri)
            pf=PROPFIND(uri,ic,d)

            if request.body:
                pf.read_propfind(request.body)

            try:
                DATA=pf.createResponse(host)
                DATA=DATA+"\n\n"
                DATA=string.replace(DATA,"<D:response>","<D:response xmlns:D='DAV:'>")
            except DAV_Forbidden:
                return self._sendAuthentication(self.authenticationRealm)
            except DAV_Error, (ec,dd):
                return self.create_status(ec)
            return self.create_body_chunks(DATA,"207","Multi-Status","Multiple responses")
        finally:
            repo.txRollback()



    def do_GET(self,request):
        """Serve a GET request."""

        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error
        try:

            host = self._getHost(repo, request)
            uri = request.path[1:]

            ic = self.interfaceClass()
            uri=urlparse.urljoin(host.getUri(),request.uri)
            uri=urllib.unquote(uri)

            # get the last modified date
            try:
                lm=ic.get_prop(uri,"DAV:","getlastmodified")
            except:
                lm="Sun, 01 Dec 2014 00:00:00 GMT"  # dummy!
            headers={"Last-Modified":lm}

            # get the content type
            try:
                ct=ic.get_prop(uri,"DAV:","getcontenttype")
            except:
                ct="application/octet-stream"

            # get the data
            try:
                data=ic.get_data(uri,host)
            except DAV_Forbidden:
                return self._sendAuthentication(self.authenticationRealm)
            except DAV_Error, (ec,dd):
                return self.create_status(ec)

            # send the data
            return self.create_body(data,"200","OK","OK",ct,headers)
        finally:
            repo.txRollback()

    def do_POST(self,request):
        return self._createErrorResponse(request,404,uri=request.uri)

    def do_HEAD(self,request):
        res = self.do_GET(request)
        res.body = ''
        return res


    def do_MKCOL(self,request):
        """ create a new collection """
        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error
        try:

            commit = 0
            host = self._getHost(repo,request)
            ic = self.interfaceClass()
            uri=urlparse.urljoin(host.getUri(), request.uri)
            uri=urllib.unquote(uri)

            try:
                ic.mkcol(uri,host)
            except DAV_Forbidden:
                return self._sendAuthentication(self.authenticationRealm)
            except DAV_Error, (ec,dd):
                return self.create_status(ec)
            commit = 1
            return self.create_status(200)
        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()


    def do_DELETE(self,request):

        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error
        try:

            commit = 0
            host = self._getHost(repo,request)

            ic = self.interfaceClass()
            uri=urlparse.urljoin(host.getUri(), request.uri)
            uri=urllib.unquote(uri)


            dl=DELETE(uri,ic)
            try:
                if ic.is_collection(uri,host):
                    res=dl.delcol(host)
                else:
                    res=dl.delone(host)
            except DAV_Forbidden:
                return self._sendAuthentication(self.authenticationRealm)
            except DAV_Error, (ec,dd):
                return self.create_status(ec)

            commit = 1
            if res:
                return self.create_status(207,body=res)
            else:
                return self.create_status(204)

        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()


    def do_PUT(self,request):

        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error
        try:

            commit = 0
            host = self._getHost(repo,request)

            ic = self.interfaceClass()
            uri=urlparse.urljoin(host.getUri(), request.uri)
            uri=urllib.unquote(uri)




            # read the body
            ct='text/plain'
            if request.headers.has_key("Content-Type"):
                ct=request.headers['Content-Type']

            try:
                ic.put(uri,request.body,ct,host)
            except DAV_Error, (ec,dd):
                return self.create_status(ec)
            commit = 1
            return self.create_status(201)
        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()


    def do_COPY(self,request):
    """ copy one resource to another """
        return self.copymove(COPY,request)

    def do_MOVE(self,request):
    """ move one resource to another """
        return self.copymove(MOVE,request)

    def copymove(self,CLASS,request):
    """ common method for copying or moving objects """

        rt,data = self._do_authentication(request.headers,self.authenticationRealm)
        if not rt:
            return data
        sid = data
        tx = CosTransactionsLib.Create()
        try:

            commit = 0
            mxs = MetaXmlServerLib.GetMetaRepository(tx,
                                                     sid,
                                                     xmlServerBaseUri = self.xmlServerBaseUri,
                                                     rdfModelUri = self.rdfModelUri,
                                                     serverName = self.serverName)

            ic = self.interfaceClass()

            # get the source URI
            source_uri=urlparse.urljoin(self.xmlServerBaseUri,request.uri)
            source_uri=urllib.unquote(source_uri)

            # get the destination URI
            dest_uri=request.headers['Destination']
            dest_uri=urllib.unquote(dest_uri)
            #Give it the same host
            uparts=urlparse.urlparse(dest_uri)
            dest_uri=urlparse.urljoin(self.xmlServerBaseUri,uparts[2])

            # Overwrite?
            overwrite=1
            result_code=204
            if request.headers.has_key("Overwrite"):
                if request.headers['Overwrite']=="F":
                    overwrite=None
                    result_code=201

            # instanciate ACTION class
            cp=CLASS(ic,source_uri,dest_uri,overwrite)

            # Depth?
            d="infinity"
            if request.headers.has_key("Depth"):
                d=request.headers['Depth']
                if d!="0" and d!="infinity":
                    return self.create_status(400)
                if d=="0":
                    res=cp.single_action(mxs)
                    return self.create_status(res)

            # now it only can be "infinity" but we nevertheless check for a collection
            if ic.is_collection(source_uri,mxs):
                try:
                    res=cp.tree_action(mxs)
                except DAV_Error, (ec,dd):
                    return self.create_status(ec)
            else:
                try:
                    res=cp.single_action(mxs)
                except DAV_Error, (ec,dd):
                    return self.create_status(ec)

            commit = 1
            if res:
        return self.create_body_chunks(res,207,STATUS_CODES[207],STATUS_CODES[207],
                                               ctype='text/xml; charset="utf-8"')
            else:
            return self.create_status(result_code)
        finally:
            UserServerLib.Logout(sid)
            if commit:
                tx.get_terminator().commit(0)
            else:
                tx.get_terminator().rollback()




    def do_LOCK(self,request):
        body="""<?xml version="1.0" encoding="utf-8" ?>
         <D:prop xmlns:D="DAV:">
           <D:lockdiscovery>
                <D:activelock>
                     <D:locktype><D:write/></D:locktype>
                     <D:lockscope><D:exclusive/></D:lockscope>
                     <D:depth>Infinity</D:depth>
                     <D:owner>
                          <D:href>molson</D:href>
                     </D:owner>
                     <D:timeout>Second-604800</D:timeout>
                     <D:locktoken>
                          <D:href>
                     opaquelocktoken:e71d4fae-5dec-22d6-fea5-00a0c91e6be4
                          </D:href>
                     </D:locktoken>
                </D:activelock>
           </D:lockdiscovery>
         </D:prop>"""

        return self.create_body_chunks(body,207,STATUS_CODES[207],STATUS_CODES[207],
                                       ctype='text/xml; charset="utf-8"')

    def do_PROPPATCH(self,request):
        return self._sendError(404,"File not found",{},'')



    def do_TRACE(self,request):
        return self._sendError(404,"File not found",{},'')


    def create_status(self,code=200,mediatype='text/xml;  charset="utf-8"', \
                                msg=None,body=''):
        if not msg: msg=STATUS_CODES[code]
        return self.create_body(body,code,STATUS_CODES[code],msg,mediatype)

    def create_body_chunks(self,DATA,code,msg,desc,ctype='text/xml; encoding="utf-8"'):
    """ send a body in chunks """

        headers = {}
        headers["Content-Type"] =  ctype
        headers["Connection"] =  "close"
        headers["Transfer-Encoding"] = "chunked"

    body = hex(len(DATA))[2:]+"\r\n"
    body = body + DATA
    body = body + "\r\n0\r\n\r\n"

        return CommonHandler.Response(207,headers,body)

    def create_body(self,DATA,code,msg,desc,ctype='application/octet-stream',headers=None):
    """ send a body in one part """

        headers = headers or {}
        headers["Connection"] = "close"
        headers["Accept-Ranges"] = "bytes"
    if DATA:
        headers["Content-Length"] =  str(len(DATA))
        headers["Content-Type"] =  ctype
    else:
        headers["Content-Length"] = "0"

        return CommonHandler.Response(code,headers,DATA)

    def _getHost(self, repo, request):
        path = self.documentRoot or request.hostName or '/'
        return repo.fetchContainer(path)


def Register(props):
    # Initial retrieval and caching of alias list

    h = FssDavHandler(FourSSXmlServerDAVProxy,
                      props['DocumentRoot'],
                      props['AuthenticationRealm'],
                      )

    return [(h.do_PROPFIND, 'PROPFIND'),
            (h.do_GET, 'GET'),
            (h.do_POST, 'POST'),
            (h.do_HEAD, 'HEAD'),
            (h.do_COPY, 'COPY'),
            (h.do_MOVE, 'MOVE'),
            (h.do_PUT, 'PUT'),
            (h.do_PROPPATCH, 'PROPPATCH'),
            (h.do_MKCOL, 'MKCOL'),
            (h.do_DELETE, 'DELETE'),
            (h.do_TRACE, 'TRACE'),
            (h.do_LOCK, 'LOCK'),
            ]


